==> cds-services-impl-2.8.1-sources.jar: com/sap/cds/services/impl/authorization/AuthorizationDefaultOnHandler.java <==
/**************************************************************************
 * (C) 2019-2024 SAP SE or an SAP affiliate company. All rights reserved. *
 **************************************************************************/
package com.sap.cds.services.impl.authorization;

import java.util.List;
import java.util.stream.Collectors;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.sap.cds.ql.CQL;
import com.sap.cds.ql.Select;
import com.sap.cds.ql.cqn.CqnPredicate;
import com.sap.cds.ql.cqn.CqnSelect;
import com.sap.cds.reflect.CdsModel;
import com.sap.cds.services.authorization.ActionAccessEventContext;
import com.sap.cds.services.authorization.AuthorizationService;
import com.sap.cds.services.authorization.CalcWhereConditionEventContext;
import com.sap.cds.services.authorization.EntityAccessEventContext;
import com.sap.cds.services.authorization.FunctionAccessEventContext;
import com.sap.cds.services.authorization.GetRestrictionEventContext;
import com.sap.cds.services.authorization.ServiceAccessEventContext;
import com.sap.cds.services.changeset.ChangeSetContextSPI;
import com.sap.cds.services.handler.EventHandler;
import com.sap.cds.services.handler.annotations.HandlerOrder;
import com.sap.cds.services.handler.annotations.On;
import com.sap.cds.services.handler.annotations.ServiceName;
import com.sap.cds.services.impl.authorization.PredicateResolver.MultipleAttributeValuesNotSupportedException;
import com.sap.cds.services.impl.draft.ParentEntityLookup;
import com.sap.cds.services.impl.utils.CdsServiceUtils;
import com.sap.cds.services.persistence.PersistenceService;
import com.sap.cds.services.request.RequestContext;
import com.sap.cds.services.runtime.CdsRuntime;
import com.sap.cds.services.utils.CdsErrorStatuses;
import com.sap.cds.services.utils.ErrorStatusException;
import com.sap.cds.services.utils.OrderConstants;
import com.sap.cds.services.utils.StringUtils;
import com.sap.cds.services.utils.TenantAwareCache;
import com.sap.cds.services.utils.model.Privilege;
import com.sap.cds.services.utils.model.Restriction;

@ServiceName(value = "*", type = AuthorizationService.class)
public class AuthorizationDefaultOnHandler implements EventHandler {

	private final static Logger logger = LoggerFactory.getLogger(AuthorizationDefaultOnHandler.class);

	private final TenantAwareCache<RestrictionLookup, CdsModel> restrictionLookupCache;
	private final TenantAwareCache<PredicateLookup, CdsModel> predicateLookupCache;
	private final TenantAwareCache<ParentEntityLookup, CdsModel> parentEntityLookups;

	private final boolean isEmptyAttributeValuesAreRestricted;
	
	AuthorizationDefaultOnHandler(CdsRuntime runtime) {
		restrictionLookupCache = TenantAwareCache.create(() -> new RestrictionLookup(), runtime);
		predicateLookupCache = TenantAwareCache.create(() -> new PredicateLookup(), runtime);
		parentEntityLookups = TenantAwareCache.create(() -> new ParentEntityLookup(RequestContext.getCurrent(runtime).getModel()), runtime);
		
		isEmptyAttributeValuesAreRestricted = runtime.getEnvironment().getCdsProperties().getSecurity().getAuthorization().isEmptyAttributeValuesAreRestricted();
	}

	@On
	@HandlerOrder(OrderConstants.On.DEFAULT_ON)
	protected void defaultGetRestriction(GetRestrictionEventContext context) {

		Restriction restriction = null;

		switch(context.getKind()) {
		case SERVICE:
			restriction = restrictionLookupCache.findOrCreate().retrieveServiceRestriction(context.getModel(), context.getName());
			break;
		case ENTITY:
			restriction = restrictionLookupCache.findOrCreate().retrieveEntityRestriction(context.getModel(), context.getName());
			break;
		case ACTION:
			restriction = restrictionLookupCache.findOrCreate().lookupActionRestriction(context.getModel(), context.getName(), context.getEventName());
			break;
		case FUNCTION:
			restriction = restrictionLookupCache.findOrCreate().lookupFunctionRestriction(context.getModel(), context.getName(), context.getEventName());
			break;
		default:
			throw new ErrorStatusException(CdsErrorStatuses.UNSUPPORTED_RESTRICTION, context.getName(), context.getKind());
		}

		context.setResult(restriction);
	}

	@On
	@HandlerOrder(OrderConstants.On.DEFAULT_ON)
	protected void defaultHasServiceAccess(ServiceAccessEventContext context) {

		final String event = context.getAccessEventName();
		final String service = context.getAccessServiceName();

		boolean result = true; // as long as there is no restriction access is granted

		Restriction restriction = RestrictionUtils.getServiceRestriction(context.getService(), service, event);
		if (restriction != null) {
			result = RestrictionUtils.passesRestriction(restriction, context.getUserInfo(), event);
		}

		context.setResult(result);
	}

	@On
	@HandlerOrder(OrderConstants.On.DEFAULT_ON)
	protected void defaultHasEntityAccess(EntityAccessEventContext context) {

		boolean result;

		ChangeSetContextSPI changeSetContext = (ChangeSetContextSPI) context.getChangeSetContext();
		// TODO adapt this, if authorization handler starts to support multiple persistence services
		if (changeSetContext == null || !changeSetContext.hasChangeSetMember(PersistenceService.DEFAULT_NAME)) {
			// Open own temporary ChangeSetContext which we don't want to left open (custom handlers might trigger remote calls)
			result = context.getCdsRuntime().changeSetContext().run(c -> {
				return defaultHasEntityAccessImpl(context);
			});
		} else {
			result = defaultHasEntityAccessImpl(context);
		}

		context.setResult(result);
	}

	protected boolean defaultHasEntityAccessImpl(EntityAccessEventContext context) {

		final String event = context.getAccessEventName();
		final String targetEntityName = context.getAccessEntityName();

		AuthorizationEntity authEntity = AuthorizationEntity.resolve(context,
				entity -> parentEntityLookups.findOrCreate().lookupParent(entity));

		if (authEntity.getAuthorizationEntity() == null) {
			// no "authorization entity" found -> reject as direct access to auto-exposed entity not allowed
			logger.debug("No authorization entity found when sending event '{}' to entity '{}'", event, targetEntityName);
			return false;
		}

		final String authorizationEntityName = authEntity.getAuthorizationEntity().getQualifiedName();

		// check static access to authorization entity in any case (target or path)
		Restriction restriction = RestrictionUtils.getEntityRestriction(context.getService(), authorizationEntityName, event);
		if (restriction != null && !RestrictionUtils.passesRestriction(restriction, context.getUserInfo(), event)) {
			if (authorizationEntityName.equals(targetEntityName)) {
				logger.debug("No authorization to send event '{}' to entity '{}'",
						event, targetEntityName);
			} else {
				logger.debug("No authorization to send event '{}' to entity '{}' because of authorization entity '{}'",
						event, targetEntityName, authorizationEntityName);
			}
			return false;
		}

		// check instance-based condition on path if required
		// We can't test by sending READ on the service level as 'event' implies a different conditions in general.
		// Instead we take the condition of 'event' and test the visibility of the instance with the condition applied.
		// Note on drafts: the authorization is enforced by DraftHandler which will filter all draft documents with the correct owner.
		if ( !authorizationEntityName.equals(targetEntityName) && authEntity.getAuthorizationPath() != null && !authEntity.isDraft()) {
			CqnPredicate authorizationCondition = authEntity.authorizationCondition(authEntity.getAuthorizationEntity(), event);
			if (authorizationCondition != null) {
				CqnSelect readTest = Select.from(authEntity.getAuthorizationPath()).where(authorizationCondition);
				// TODO: This makes generic authorization handler dependent from persistence service, but sending to the
				// corresponding application server is not correct (generic and custom handlers would be called with wrong event 'READ').
				// We decided to tackle this general problem when we start to support several persistence service instances.
				if ( !CdsServiceUtils.getDefaultPersistenceService(context).run(readTest).first().isPresent() ) {
					logger.debug("No authorization to send event '{}' to entity '{}' because of instance '{}'",
							event, targetEntityName, authEntity.getAuthorizationPath());
					return false;
				}
			}
		}

		// no further restrictions
		return true;
	}

	@On
	@HandlerOrder(OrderConstants.On.DEFAULT_ON)
	protected void defaultHasFunctionAccess(FunctionAccessEventContext context) {

		final String functionName = context.getFunctionName();
		final String entityName = context.getEntityName();

		boolean result = true; // as long as there is no restriction access is granted

		Restriction restriction = RestrictionUtils.getFunctionRestriction(context.getService(), entityName, functionName);
		if (restriction != null) {
			result = RestrictionUtils.passesRestriction(restriction, context.getUserInfo(), Privilege.PredefinedGrant.ALL.toString() /* = event: restrictions on functions have implicitly grant '*' */);
		}

		context.setResult(result);
	}

	@On
	@HandlerOrder(OrderConstants.On.DEFAULT_ON)
	protected void defaultHasActionAccess(ActionAccessEventContext context) {

		final String actionName = context.getActionName();
		final String entityName = context.getEntityName();

		boolean result = true; // as long as there is no restriction access is granted

		Restriction restriction = RestrictionUtils.getActionRestriction(context.getService(), entityName, actionName);
		if (restriction != null) {
			result = RestrictionUtils.passesRestriction(restriction, context.getUserInfo(), Privilege.PredefinedGrant.ALL.toString() /* = event: restrictions on actions have implicitly grant '*' */);
		}

		context.setResult(result);
	}

	@On
	@HandlerOrder(OrderConstants.On.DEFAULT_ON)
	protected void defaultCalcWhereCondition(CalcWhereConditionEventContext context) {

		final String event = context.getEventName();
		final String entityName = context.getEntityName();

		CqnPredicate result = null; // no filter as long there is no where condition hit

		if ( !context.getUserInfo().isPrivileged() ) {
			// we can't add where clauses to CREATE / UPSERT events!
			Restriction restriction = RestrictionUtils.getEntityRestriction(context.getService(), entityName, event);
			if (restriction != null) {
				List<Privilege> privileges = RestrictionUtils.passingPrivilegesOfRestriction(restriction, context.getUserInfo(), event).collect(Collectors.toList());
				for (Privilege privilege : privileges) {
					CqnPredicate cqnPredicate = null;
					if (privilege.hasWhere()) {
						try {
							logger.debug("Resolving 'where' condition '{}'", privilege.getCxnWhereCondition());
							cqnPredicate = predicateLookupCache.findOrCreate().resolvePredicate(privilege, context.getUserInfo(), isEmptyAttributeValuesAreRestricted);

						} catch(MultipleAttributeValuesNotSupportedException e) {
							logger.debug("No authorization to send event '{}' to entity '{}' because user {} has multiple values for attribute '{}' (filter resource '{}')",
									event, entityName, context.getUserInfo().getName(), e.getAttributeName(), e.getResourceName(), e);
							throw new ErrorStatusException(CdsErrorStatuses.EVENT_FORBIDDEN_UNSUPPORTED_USER_ATTRIBUTES, event, entityName, e.getAttributeName(), e);

						} catch (Exception e) { // NOSONAR
							throw new ErrorStatusException(CdsErrorStatuses.INVALID_WHERE_CONDITION, privilege.getCxnWhereCondition(), entityName, event, context.getUserInfo().getName(), e);
						}
					} else if( !StringUtils.isEmpty(privilege.getWhereCQL()) ) {
						// no _where property (Cxn) generated for where clause, e.g. compilation error silently ignored
						throw new ErrorStatusException(CdsErrorStatuses.INCONSISTENT_WHERE_CONDITION, privilege.getWhereCQL(), entityName);
					}

					if (cqnPredicate == null) { // null evaluates to 'true'
						result = null;
						break;
					}
					result = (result != null) ? CQL.or(result, cqnPredicate) : cqnPredicate;
				}
			}
		}

		context.setResult(result);
	}

}

==> cds-services-impl-2.10.0-sources.jar: com/sap/cds/services/impl/authorization/AuthorizationDefaultOnHandler.java <==
/**************************************************************************
 * (C) 2019-2024 SAP SE or an SAP affiliate company. All rights reserved. *
 **************************************************************************/
package com.sap.cds.services.impl.authorization;

import java.util.List;
import java.util.stream.Collectors;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.sap.cds.ql.CQL;
import com.sap.cds.ql.Select;
import com.sap.cds.ql.cqn.CqnPredicate;
import com.sap.cds.ql.cqn.CqnSelect;
import com.sap.cds.reflect.CdsModel;
import com.sap.cds.services.authorization.ActionAccessEventContext;
import com.sap.cds.services.authorization.AuthorizationService;
import com.sap.cds.services.authorization.CalcWhereConditionEventContext;
import com.sap.cds.services.authorization.EntityAccessEventContext;
import com.sap.cds.services.authorization.FunctionAccessEventContext;
import com.sap.cds.services.authorization.GetRestrictionEventContext;
import com.sap.cds.services.authorization.ServiceAccessEventContext;
import com.sap.cds.services.changeset.ChangeSetContextSPI;
import com.sap.cds.services.handler.EventHandler;
import com.sap.cds.services.handler.annotations.HandlerOrder;
import com.sap.cds.services.handler.annotations.On;
import com.sap.cds.services.handler.annotations.ServiceName;
import com.sap.cds.services.impl.authorization.PredicateResolver.MultipleAttributeValuesNotSupportedException;
import com.sap.cds.services.impl.draft.ParentEntityLookup;
import com.sap.cds.services.impl.utils.CdsServiceUtils;
import com.sap.cds.services.persistence.PersistenceService;
import com.sap.cds.services.request.RequestContext;
import com.sap.cds.services.runtime.CdsRuntime;
import com.sap.cds.services.utils.CdsErrorStatuses;
import com.sap.cds.services.utils.ErrorStatusException;
import com.sap.cds.services.utils.OrderConstants;
import com.sap.cds.services.utils.StringUtils;
import com.sap.cds.services.utils.TenantAwareCache;
import com.sap.cds.services.utils.model.Privilege;
import com.sap.cds.services.utils.model.Restriction;

@ServiceName(value = "*", type = AuthorizationService.class)
public class AuthorizationDefaultOnHandler implements EventHandler {

	private final static Logger logger = LoggerFactory.getLogger(AuthorizationDefaultOnHandler.class);

	private final TenantAwareCache<RestrictionLookup, CdsModel> restrictionLookupCache;
	private final TenantAwareCache<PredicateLookup, CdsModel> predicateLookupCache;
	private final TenantAwareCache<ParentEntityLookup, CdsModel> parentEntityLookups;

	private final boolean isEmptyAttributeValuesAreRestricted;
	
	AuthorizationDefaultOnHandler(CdsRuntime runtime) {
		restrictionLookupCache = TenantAwareCache.create(() -> new RestrictionLookup(), runtime);
		predicateLookupCache = TenantAwareCache.create(() -> new PredicateLookup(), runtime);
		parentEntityLookups = TenantAwareCache.create(() -> new ParentEntityLookup(RequestContext.getCurrent(runtime).getModel()), runtime);
		
		isEmptyAttributeValuesAreRestricted = runtime.getEnvironment().getCdsProperties().getSecurity().getAuthorization().isEmptyAttributeValuesAreRestricted();
	}

	@On
	@HandlerOrder(OrderConstants.On.DEFAULT_ON)
	protected void defaultGetRestriction(GetRestrictionEventContext context) {

		Restriction restriction = null;

		switch(context.getKind()) {
		case SERVICE:
			restriction = restrictionLookupCache.findOrCreate().retrieveServiceRestriction(context.getModel(), context.getName());
			break;
		case ENTITY:
			restriction = restrictionLookupCache.findOrCreate().retrieveEntityRestriction(context.getModel(), context.getName());
			break;
		case ACTION:
			restriction = restrictionLookupCache.findOrCreate().lookupActionRestriction(context.getModel(), context.getName(), context.getEventName());
			break;
		case FUNCTION:
			restriction = restrictionLookupCache.findOrCreate().lookupFunctionRestriction(context.getModel(), context.getName(), context.getEventName());
			break;
		default:
			throw new ErrorStatusException(CdsErrorStatuses.UNSUPPORTED_RESTRICTION, context.getName(), context.getKind());
		}

		context.setResult(restriction);
	}

	@On
	@HandlerOrder(OrderConstants.On.DEFAULT_ON)
	protected void defaultHasServiceAccess(ServiceAccessEventContext context) {

		final String event = context.getAccessEventName();
		final String service = context.getAccessServiceName();

		boolean result = true; // as long as there is no restriction access is granted

		Restriction restriction = RestrictionUtils.getServiceRestriction(context.getService(), service, event);
		if (restriction != null) {
			result = RestrictionUtils.passesRestriction(restriction, context.getUserInfo(), event);
		}

		context.setResult(result);
	}

	@On
	@HandlerOrder(OrderConstants.On.DEFAULT_ON)
	protected void defaultHasEntityAccess(EntityAccessEventContext context) {

		boolean result;

		ChangeSetContextSPI changeSetContext = (ChangeSetContextSPI) context.getChangeSetContext();
		// TODO adapt this, if authorization handler starts to support multiple persistence services
		if (changeSetContext == null || !changeSetContext.hasChangeSetMember(PersistenceService.DEFAULT_NAME)) {
			// Open own temporary ChangeSetContext which we don't want to left open (custom handlers might trigger remote calls)
			result = context.getCdsRuntime().changeSetContext().run(c -> {
				return defaultHasEntityAccessImpl(context);
			});
		} else {
			result = defaultHasEntityAccessImpl(context);
		}

		context.setResult(result);
	}

	protected boolean defaultHasEntityAccessImpl(EntityAccessEventContext context) {

		final String event = context.getAccessEventName();
		final String targetEntityName = context.getAccessEntityName();

		AuthorizationEntity authEntity = AuthorizationEntity.resolve(context,
				entity -> parentEntityLookups.findOrCreate().lookupParent(entity));

		if (authEntity.getAuthorizationEntity() == null) {
			// no "authorization entity" found -> reject as direct access to auto-exposed entity not allowed
			logger.debug("No authorization entity found when sending event '{}' to entity '{}'", event, targetEntityName);
			return false;
		}

		final String authorizationEntityName = authEntity.getAuthorizationEntity().getQualifiedName();

		// check static access to authorization entity in any case (target or path)
		Restriction restriction = RestrictionUtils.getEntityRestriction(context.getService(), authorizationEntityName, event);
		if (restriction != null && !RestrictionUtils.passesRestriction(restriction, context.getUserInfo(), event)) {
			if (authorizationEntityName.equals(targetEntityName)) {
				logger.debug("No authorization to send event '{}' to entity '{}'",
						event, targetEntityName);
			} else {
				logger.debug("No authorization to send event '{}' to entity '{}' because of authorization entity '{}'",
						event, targetEntityName, authorizationEntityName);
			}
			return false;
		}

		// check instance-based condition on path if required
		// We can't test by sending READ on the service level as 'event' implies a different conditions in general.
		// Instead we take the condition of 'event' and test the visibility of the instance with the condition applied.
		// Note on drafts: the authorization is enforced by DraftHandler which will filter all draft documents with the correct owner.
		if ( !authorizationEntityName.equals(targetEntityName) && authEntity.getAuthorizationPath() != null && !authEntity.isDraft()) {
			CqnPredicate authorizationCondition = authEntity.authorizationCondition(authEntity.getAuthorizationEntity(), event);
			if (authorizationCondition != null) {
				CqnSelect readTest = Select.from(authEntity.getAuthorizationPath()).where(authorizationCondition);
				// TODO: This makes generic authorization handler dependent from persistence service, but sending to the
				// corresponding application server is not correct (generic and custom handlers would be called with wrong event 'READ').
				// We decided to tackle this general problem when we start to support several persistence service instances.
				if ( CdsServiceUtils.getDefaultPersistenceService(context).run(readTest).first().isEmpty() ) {
					logger.debug("No authorization to send event '{}' to entity '{}' because of instance '{}'",
							event, targetEntityName, authEntity.getAuthorizationPath());
					return false;
				}
			}
		}

		// no further restrictions
		return true;
	}

	@On
	@HandlerOrder(OrderConstants.On.DEFAULT_ON)
	protected void defaultHasFunctionAccess(FunctionAccessEventContext context) {

		final String functionName = context.getFunctionName();
		final String entityName = context.getEntityName();

		boolean result = true; // as long as there is no restriction access is granted

		Restriction restriction = RestrictionUtils.getFunctionRestriction(context.getService(), entityName, functionName);
		if (restriction != null) {
			result = RestrictionUtils.passesRestriction(restriction, context.getUserInfo(), Privilege.PredefinedGrant.ALL.toString() /* = event: restrictions on functions have implicitly grant '*' */);
		}

		context.setResult(result);
	}

	@On
	@HandlerOrder(OrderConstants.On.DEFAULT_ON)
	protected void defaultHasActionAccess(ActionAccessEventContext context) {

		final String actionName = context.getActionName();
		final String entityName = context.getEntityName();

		boolean result = true; // as long as there is no restriction access is granted

		Restriction restriction = RestrictionUtils.getActionRestriction(context.getService(), entityName, actionName);
		if (restriction != null) {
			result = RestrictionUtils.passesRestriction(restriction, context.getUserInfo(), Privilege.PredefinedGrant.ALL.toString() /* = event: restrictions on actions have implicitly grant '*' */);
		}

		context.setResult(result);
	}

	@On
	@HandlerOrder(OrderConstants.On.DEFAULT_ON)
	protected void defaultCalcWhereCondition(CalcWhereConditionEventContext context) {

		final String event = context.getEventName();
		final String entityName = context.getEntityName();

		CqnPredicate result = null; // no filter as long there is no where condition hit

		if ( !context.getUserInfo().isPrivileged() ) {
			// we can't add where clauses to CREATE / UPSERT events!
			Restriction restriction = RestrictionUtils.getEntityRestriction(context.getService(), entityName, event);
			if (restriction != null) {
				List<Privilege> privileges = RestrictionUtils.passingPrivilegesOfRestriction(restriction, context.getUserInfo(), event).collect(Collectors.toList());
				for (Privilege privilege : privileges) {
					CqnPredicate cqnPredicate = null;
					if (privilege.hasWhere()) {
						try {
							logger.debug("Resolving 'where' condition '{}'", privilege.getCxnWhereCondition());
							cqnPredicate = predicateLookupCache.findOrCreate().resolvePredicate(privilege, context.getUserInfo(), isEmptyAttributeValuesAreRestricted);

						} catch(MultipleAttributeValuesNotSupportedException e) {
							logger.debug("No authorization to send event '{}' to entity '{}' because user {} has multiple values for attribute '{}' (filter resource '{}')",
									event, entityName, context.getUserInfo().getName(), e.getAttributeName(), e.getResourceName(), e);
							throw new ErrorStatusException(CdsErrorStatuses.EVENT_FORBIDDEN_UNSUPPORTED_USER_ATTRIBUTES, event, entityName, e.getAttributeName(), e);

						} catch (Exception e) { // NOSONAR
							throw new ErrorStatusException(CdsErrorStatuses.INVALID_WHERE_CONDITION, privilege.getCxnWhereCondition(), entityName, event, context.getUserInfo().getName(), e);
						}
					} else if( !StringUtils.isEmpty(privilege.getWhereCQL()) ) {
						// no _where property (Cxn) generated for where clause, e.g. compilation error silently ignored
						throw new ErrorStatusException(CdsErrorStatuses.INCONSISTENT_WHERE_CONDITION, privilege.getWhereCQL(), entityName);
					}

					if (cqnPredicate == null) { // null evaluates to 'true'
						result = null;
						break;
					}
					result = (result != null) ? CQL.or(result, cqnPredicate) : cqnPredicate;
				}
			}
		}

		context.setResult(result);
	}

}
